// TrackBall.cpp: implementation of the TrackBall class.
//
/*
 * Ok, simulate a track-ball.  Project the points onto the virtual
 * trackball, then figure out the axis of rotation, which is the cross
 * product of P1 P2 and O P1 (O is the center of the ball, 0,0,0)
 * Note:  This is a deformed trackball-- is a trackball in the center,
 * but is deformed into a hyperbolic sheet of rotation away from the
 * center.  This particular function was chosen after trying out
 * several variations.
 *
 * It is assumed that the arguments to this routine are in the range
 * (-1.0 ... 1.0)
 */
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "TrackBall.h"

#include <math.h>
#include <GL/gl.h>


#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

/*
 * This size should really be based on the distance from the center of
 * rotation to the point on the object underneath the mouse.  That
 * point would then track the mouse as closely as possible.  This is a
 * simple example, though, so that is left as an Exercise for the
 * Programmer.
 */
#define TRACKBALLSIZE  (0.78375f)	// don't chage this value

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

TrackBall::TrackBall(float p1x, float p1y, float p2x, float p2y)
{
    float a[3]; /* Axis of rotation */
    float phi;  /* how much to rotate about axis */
    float p1[3], p2[3], d[3];
    float t;

    _trackBallSize = TRACKBALLSIZE;

    /* Zero rotation */
    vzero(_currentQuat);
    _currentQuat[3] = 1.0;

    if (p1x == p2x && p1y == p2y) 
   	return; 
    /*
     * First, figure out z-coordinates for projection of P1 and P2 to
     * deformed sphere
     */
    vset(p1,p1x,p1y,tb_project_to_sphere(_trackBallSize,p1x,p1y));
    vset(p2,p2x,p2y,tb_project_to_sphere(_trackBallSize,p2x,p2y));

    /*
     *  Now, we want the cross product of P1 and P2
     */
    vcross(p2,p1,a);

    /*
     *  Figure out how much to rotate around that axis.
     */
    vsub(p1,p2,d);
    t = (float) (vlength(d) / (2.0*_trackBallSize));

    /*
     * Avoid problems with out-of-control values...
     */
    if (t > 1.0) t = 1.0;
    if (t < -1.0) t = -1.0;
    phi = (float) (2.0 * asin(t));

    axis_to_quat(a,phi,_currentQuat);
    
}

TrackBall::TrackBall()
{
    _trackBallSize = TRACKBALLSIZE;

    /* Zero rotation */
    vzero(_currentQuat);
    _currentQuat[3] = 1.0;
}

void
TrackBall::vzero(float *v)
{
    v[0] = 0.0;
    v[1] = 0.0;
    v[2] = 0.0;
}

void
TrackBall::vset(float *v, float x, float y, float z)
{
    v[0] = x;
    v[1] = y;
    v[2] = z;
}

void
TrackBall::vsub(const float *src1, const float *src2, float *dst)
{
    dst[0] = src1[0] - src2[0];
    dst[1] = src1[1] - src2[1];
    dst[2] = src1[2] - src2[2];
}

void
TrackBall::vcopy(const float *v1, float *v2)
{
    register int i;
    for (i = 0 ; i < 3 ; i++)
        v2[i] = v1[i];
}

void
TrackBall::vcross(const float *v1, const float *v2, float *cross)
{
    float temp[3];

    temp[0] = (v1[1] * v2[2]) - (v1[2] * v2[1]);
    temp[1] = (v1[2] * v2[0]) - (v1[0] * v2[2]);
    temp[2] = (v1[0] * v2[1]) - (v1[1] * v2[0]);
    vcopy(temp, cross);
}

float
TrackBall::vlength(const float *v)
{
    return ((float) sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2]));
}

void
TrackBall::vscale(float *v, float div)
{
    v[0] *= div;
    v[1] *= div;
    v[2] *= div;
}

void
TrackBall::vnormal(float *v)
{
    vscale(v,1.0f/vlength(v));
}

float
TrackBall::vdot(const float *v1, const float *v2)
{
    return v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2];
}

void
TrackBall::vadd(const float *src1, const float *src2, float *dst)
{
    dst[0] = src1[0] + src2[0];
    dst[1] = src1[1] + src2[1];
    dst[2] = src1[2] + src2[2];
}

/*
 *  Given an axis and angle, compute quaternion.
 */
void
TrackBall::axis_to_quat(float a[3], float phi, float q[4])
{
    vnormal(a);
    vcopy(a,q);
    vscale(q,(float)sin(phi/2.0));
    q[3] = (float) cos(phi/2.0);
}

/*
 * Project an x,y pair onto a sphere of radius r OR a hyperbolic sheet
 * if we are away from the center of the sphere.
 */
float
TrackBall::tb_project_to_sphere(float r, float x, float y)
{
    float d, t, z;

    d = (float) sqrt(x*x + y*y);
    if (d < r * 0.70710678118654752440) {    /* Inside sphere */
        z = (float) sqrt(r*r - d*d);
    } else {           /* On hyperbola */
        t = (float) (r / 1.41421356237309504880);
        z = t*t / d;
    }
    return z;
}

/*
 * Given two rotations, e1 and e2, expressed as quaternion rotations,
 * figure out the equivalent single rotation and stuff it into dest.
 *
 * This routine also normalizes the result every RENORMCOUNT times it is
 * called, to keep error from creeping in.
 *
 * NOTE: This routine is written so that q1 or q2 may be the same
 * as dest (or each other).
 */

//#define RENORMCOUNT 97

void
TrackBall::add_quats(float q1[4])
{
    float t1[4], t2[4], t3[4];
    float tf[4];

    vcopy(q1,t1);
    vscale(t1,_currentQuat[3]);

    vcopy(_currentQuat,t2);
    vscale(t2,q1[3]);

    vcross(_currentQuat,q1,t3);
    vadd(t1,t2,tf);
    vadd(t3,tf,tf);
    tf[3] = q1[3] * _currentQuat[3] - vdot(q1,_currentQuat);

    _currentQuat[0] = tf[0];
    _currentQuat[1] = tf[1];
    _currentQuat[2] = tf[2];
    _currentQuat[3] = tf[3];

 //   normalize_quat(_currentQuat);

}

/*
 * Quaternions always obey:  a^2 + b^2 + c^2 + d^2 = 1.0
 * If they don't add up to 1.0, dividing by their magnitued will
 * renormalize them.
 *
 * Note: See the following for more information on quaternions:
 *
 * - Shoemake, K., Animating rotation with quaternion curves, Computer
 *   Graphics 19, No 3 (Proc. SIGGRAPH'85), 245-254, 1985.
 * - Pletinckx, D., Quaternion calculus as a basic tool in computer
 *   graphics, The Visual Computer 5, 2-13, 1989.
 */

void
TrackBall::normalize_quat(float q[4])
{
    int i;
    float mag;

    mag = (q[0]*q[0] + q[1]*q[1] + q[2]*q[2] + q[3]*q[3]);
    for (i = 0; i < 4; i++) q[i] /= mag;
}

/*
 * Build a rotation matrix, given a quaternion rotation.
 *
 */
void
TrackBall::build_rotmatrix()
{
	_m[0][0] = (float) (1.0 - 2.0 * (_currentQuat[1] * _currentQuat[1] + _currentQuat[2] * _currentQuat[2]));
	_m[0][1] = (float) (2.0 * (_currentQuat[0] * _currentQuat[1] - _currentQuat[2] * _currentQuat[3]));
	_m[0][2] = (float) (2.0 * (_currentQuat[2] * _currentQuat[0] + _currentQuat[1] * _currentQuat[3]));
	_m[0][3] = 0.0f;

	_m[1][0] = (float) (2.0 * (_currentQuat[0] * _currentQuat[1] + _currentQuat[2] * _currentQuat[3]));
	_m[1][1]= (float) (1.0 - 2.0 * (_currentQuat[2] * _currentQuat[2] + _currentQuat[0] * _currentQuat[0]));
	_m[1][2] = (float) (2.0 * (_currentQuat[1] * _currentQuat[2] - _currentQuat[0] * _currentQuat[3]));
	_m[1][3] = 0.0f;

	_m[2][0] = (float) (2.0 * (_currentQuat[2] * _currentQuat[0] - _currentQuat[1] * _currentQuat[3]));
	_m[2][1] = (float) (2.0 * (_currentQuat[1] * _currentQuat[2] + _currentQuat[0] * _currentQuat[3]));
	_m[2][2] = (float) (1.0 - 2.0 * (_currentQuat[1] * _currentQuat[1] + _currentQuat[0] * _currentQuat[0]));
	_m[2][3] = 0.0f;

	_m[3][0] = 0.0f;
	_m[3][1] = 0.0f;
	_m[3][2] = 0.0f;
	_m[3][3] = 1.0f;
}


void
TrackBall::setTrackBallSize(float size)
{
	_trackBallSize = size;
}


float *
TrackBall::return_quats() const 
{
	return (float *) _currentQuat;
}


void
TrackBall::getMatrix( float matrix[][4] ) 
{
	build_rotmatrix();

	for ( int i = 0 ; i < 4 ; i ++ )
	for ( int j = 0 ; j < 4 ; j ++ )
		matrix[i][j] = _m[i][j];

}

GLfloat *TrackBall::return_matrix() 
{
	build_rotmatrix();

	return (GLfloat *) _m;
}

void TrackBall::init()
{
    /* Zero rotation */
    vzero(_currentQuat);
    _currentQuat[3] = 1.0;
}
